/* 
 * SCIE -- Spinal Cord Injury Ontology Extraction
 * Copyright (C) 2013, 2014
 * Raphael Dickfelder, Jan Göpfert, Benjamin Paaßen, Andreas Stöckel
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.techfak.scie.pdfextractor.structure;

/**
 * The LineSegment class implements the AbstractLineSegmentSegment interface and
 * adds (static) utility functions that help to compare to lines.
 *
 * @author Andreas Stoeckel - astoecke(at)techfak.uni-bielefeld.de
 */
public class LineSegment implements AbstractLineSegment {

	/**
	 * Returns true if the given line is valid (its begin is smaller or equal to
	 * its end).
	 *
	 * @param line is the line that should be checked for validity.
	 * @return true if the line is valid, false otherwise.
	 */
	public static final boolean isValid(AbstractLineSegment line) {
		return (line.getBegin() <= line.getEnd());
	}

	/**
	 * Returns the length of the line segment.
	 *
	 * @param line is the line segment of which the (absolute) length is calculated.
	 * @return the absolute value of the length of the line segment.
	 */
	public static final int length(AbstractLineSegment line) {
		return Math.abs(line.getEnd() - line.getBegin());
	}

	/**
	 * Normalizes the line segment by swapping begin and end if they are in the
	 * wrong order.
	 *
	 * @param line is the abstract line segment to be normalized.
	 * @return a new, normalized LineSegment instance.
	 */
	public static final LineSegment normalize(AbstractLineSegment line) {
		if (!isValid(line)) {
			return new LineSegment(line.getEnd(), line.getBegin());
		} else {
			return new LineSegment(line.getBegin(), line.getEnd());
		}
	}

	/**
	 * Returns the boundaries of the normalized line as a two-element array --
	 * normalization means that begin and end are swapped if end is larger than
	 * begin.
	 *
	 * @param line is the line of which the (normalized) bounds are returned.
	 * @return a two-element array containing begin (0) and end (1).
	 */
	public static final int[] normalizedBounds(AbstractLineSegment line) {
		AbstractLineSegment l = normalize(line);
		return new int[]{l.getBegin(), l.getEnd()};
	}

	/**
	 * Returns a new line which is the union of the two lines.
	 *
	 * @param l1 is the fist line segment for forming the union.
	 * @param l2 is the second line segment for forming the union.
	 * @return a new line segment that forms the union
	 */
	public static final LineSegment union(AbstractLineSegment l1,
			AbstractLineSegment l2) {
		final int n1[] = normalizedBounds(l1);
		final int n2[] = normalizedBounds(l2);
		return new LineSegment(
				Math.min(n1[0], n2[0]),
				Math.max(n1[1], n2[1])
		);
	}

	/**
	 * Returns a new line which is the union of the two lines.
	 *
	 * @param l1 is the fist line segment for forming the union.
	 * @param l2 is the second line segment for forming the union.
	 * @return a new line segment that forms the union or null if the two line
	 * segments to not overlap.
	 */
	public static final LineSegment intersection(AbstractLineSegment l1,
			AbstractLineSegment l2) {
		final int n1[] = normalizedBounds(l1);
		final int n2[] = normalizedBounds(l2);
		final int b = Math.max(n1[0], n2[0]);
		final int e = Math.min(n1[1], n2[1]);
		if (b > e) {
			return null;
		}
		return new LineSegment(b, e);
	}

	/**
	 * Returns true if the boundaries of the two line segments are equal.
	 *
	 * @param l1 is the fist line segment for comparison.
	 * @param l2 is the second line segment for comparison.
	 * @return true if the boundaries are equal, false otherwise.
	 */
	public static final boolean boundariesEqual(AbstractLineSegment l1,
			AbstractLineSegment l2) {
		final int n1[] = normalizedBounds(l1);
		final int n2[] = normalizedBounds(l2);
		return (n1[0] == n2[0]) && (n1[1] == n2[1]);
	}

	/**
	 * Returns true if the boundaries of the two line segments overlap.
	 *
	 * @param l1 is the first line segment for comparison.
	 * @param l2 is the second line segment for comparison.
	 * @return true if the boundaries of the two line segments overlap,
	 * independent of their validity (begin and end may be swaped).
	 */
	public static final boolean overlaps(AbstractLineSegment l1,
			AbstractLineSegment l2) {
		final int n1[] = normalizedBounds(l1);
		final int n2[] = normalizedBounds(l2);
		return ((n1[0] >= n2[0]) && (n1[0] <= n2[1])) || ((n2[0] >= n1[0]) && (n2[0] <= n1[1]));
	}

	/**
	 * Start position of the line.
	 */
	protected int begin;

	/**
	 * End position of the line.
	 */
	protected int end;

	/**
	 * Initializes the line segment as invalid, with begin being set to INF and
	 * end being set to -INF.
	 */
	public LineSegment() {
		this(Integer.MAX_VALUE, Integer.MIN_VALUE);
	}

	/**
	 * Initializes the line segment with the given begin and end.
	 *
	 * @param begin is the begin.
	 * @param end is then end.
	 */
	public LineSegment(int begin, int end) {
		this.begin = begin;
		this.end = end;
	}

	/**
	 * Returns the begin position of the line.
	 *
	 * @return the begin position of the line.
	 */
	@Override
	public final int getBegin() {
		return begin;
	}

	/**
	 * Returns the end position of the line.
	 *
	 * @return the end position of the line.
	 */
	@Override
	public final int getEnd() {
		return end;
	}

	/**
	 * Sets the start position of the line.
	 *
	 * @param begin is the new start position of the line.
	 */
	public final void setBegin(int begin) {
		this.begin = begin;
	}

	/**
	 * Sets the end position of the line.
	 *
	 * @param end is the new end position of the line.
	 */
	public final void setEnd(int end) {
		this.end = end;
	}

	/**
	 * Returns true if this line is valid (its begin is smaller or equal to its
	 * end).
	 *
	 * @return true if the line is valid, false otherwise.
	 */
	public final boolean isValid() {
		return isValid(this);
	}

	/**
	 * Returns true if the boundaries of the two line segments are equal.
	 *
	 * @param l is the line segment for comparison.
	 * @return true if the boundaries are equal, false otherwise.
	 */
	public final boolean boundariesEqual(AbstractLineSegment l) {
		return boundariesEqual(this, l);
	}

	/**
	 * Returns true if the boundaries of the two line segments overlap.
	 *
	 * @param l is the line segment for comparison.
	 * @return true if the boundaries of the two line segments overlap,
	 * independent of their validity (begin and end may be swaped).
	 */
	public final boolean overlaps(AbstractLineSegment l) {
		return overlaps(this, l);
	}

	/**
	 * Swaps begin and end if the line is not valid.
	 */
	public final void normalize() {
		if (!isValid()) {
			final int tmp = end;
			end = begin;
			begin = tmp;
		}
	}

	/**
	 * Returns the length of the line segment.
	 *
	 * @return the absolute value of the length of the line segment.
	 */
	public final int length() {
		return length(this);
	}

}
